/* shar.c  11-Apr-89  A.J.Travis */

/*
 * Unix style 'shell' archiver
 */

#include <stdio.h>

#if MSDOS
#include <dos.h>
#define EOL '\n'
struct FIND *files;

#else
#define EOL '\r'
#endif

#define UNDEFINED NULL			/* i/o stream status */

#define ERRTERM		"shar: parse error - \"<<\" expected\n"
#define ERRTO		"shar: parse error - \">\" expected\n"
#define ESCRIPT		"shar: sed script not recognised\n"
#define ELINE		"shar: error extracting %s\n", path

#define TRUE		1
#define FALSE		0
#define SAME		0		/* strncmp() strings equal */
#define LINESIZE	BUFSIZ

FILE *in;				/* input file */
FILE *out;				/* output file */
char line[LINESIZE];			/* i/o buffer */
char path[80];				/* pathname */
char cmd[80];				/* sed or cat command string */
char script[80];			/* sed edit script */
char term[80];				/* 'here' << document terminator */
int skip;				/* length of shar prefix to skip */
int terml;				/* length of terminator */
int sed;				/* sed or cat command */
int cflag;				/* create archive */
int xflag;				/* extract files */

int main(argc, argv)
int argc;
char *argv[];
{
	in = UNDEFINED;			/* default input stream */
	out = UNDEFINED;		/* default output stream */
	cflag = FALSE;
	xflag = FALSE;

	/* first argument is mandatory flag */
	if (argc < 3)
		usage();
	if (strcmp(argv[1], "-a") == SAME)
		cflag++;
	else if (strcmp(argv[1], "-x") == SAME)
		xflag++;
	else
		usage();
	--argc;
	argv++;

	/* next argument is name of archive */
	if (xflag) {
		if ((in = fopen(argv[1], "rb")) == NULL)
			cant(argv[1]);

		/* for simplicity, just extract the lot */
		extract();
	}
	else {
		if ((out = fopen(argv[1], "w")) == NULL)
			cant(argv[1]);
		--argc;
		argv++;

		/* remaining arguments are files to archive */
		while (--argc) {
			archive(argv[1]);
			argv++;			
		}
	}
	exit(0);
}

/*
 * report correct usage, and exit
 */
usage()
{
	fprintf(stderr, "usage: shar [-a] [-x] archive file1 ... [filen]\n");
	exit(-1);
}

/*
 * print file access error, and exit
 */
cant(name)
char *name;				/* name of problem file */
{
	fprintf(stderr, "shar: can't open %s\n", name);
	if (in != UNDEFINED)
		close(in);
	if (out != UNDEFINED & out != stdout)
		close(out);
	exit(-1);
}

/*
 * create shell archive
 */
archive(name)
char *name;				/* name of file to archive */
{
	char *bp;			/* buffer pointer */

	if ((in = fopen(name, "r")) == NULL)
		cant(name);
	fprintf(out, "sed 's/^X//' << 'SHAR_EOF' > %s%c", name, EOL);
	while (fgets(line, LINESIZE, in) != NULL) {
		bp = line;
		while (*bp) {
			if ((*bp == '\r') | (*bp == '\n'))
				break;
			else
				bp++;
		}
		*bp++ = EOL;
		*bp = '\0';
		fprintf(out, "X%s", line);
	}
	fprintf(out, "SHAR_EOF%c", EOL);
	fclose(in);
}

/*
 * extract files from shell archive
 */
extract()
{
	int filen;
	char *p;
	char *dp;

	filen = 0;
	while (fgets(line, LINESIZE, in) != NULL) {

		/* strip shell escape chars and quotes */
		p = line;
		dp = cmd;
		while (*p) {
			if ((*p == '\\') | (*p == '\''))
				p++;
			else
				*dp++ = *p++;
		}
		*--dp = '\0';
#if MSDOS
		/* check for sub-directory tree */
		if (strncmp(cmd, "cd", 2) == SAME) {
			printf("%s\n", cmd);
			p = cmd + 2;
			dp = path;
			while (isspace(*p))
				p++;
			while (isspace(*dp++ = *p++) == FALSE)
				;
			*--dp = '\0';
			if ((files = findfirst(path, 0x10)) == NULL) {
				printf("mkdir %s\n", path);
				mkdir(path);
			}
			if (chdir(path) == -1) {
				fprintf(stderr, "shar: %s not found\n", path);
				exit(-1);
			}
			continue;
		}
#endif
		/* parse 'sed' command */
		if (strncmp(cmd, "sed", 3) == SAME) {
			sed = TRUE;
			printf("%s\n", cmd);

			/* extract 'sed' script */
			p = cmd + 3;
			dp = script;
			while (isspace(*p))
				p++;
			if (strncmp(p, "s/", 2) != SAME) {
				fprintf(stderr, ESCRIPT);
				exit(-1);
			}
			p = p + 2;
			if (*p == '^')
				p++;
			for (skip = 0; *p != '/'; skip++) {
				if ((*dp++ = *p++) == '\0') {
					fprintf(stderr, ESCRIPT);
					exit(-1);
				}
			}
			*dp = '\0';
		}

		/* parse 'cat' command */
		else if (strncmp(cmd, "cat", 3) == SAME) {
			sed = FALSE;
			printf("%s\n", cmd);
			p = cmd + 3;

		}
		else
			continue;

		/* look for "<< terminator" in command line */
		p = cmd;
		dp = term;
		while (*p++ != '<')
			;
		if (*p++ != '<') {
			fprintf(stderr, ERRTERM);
			exit(-1);
		}
		while (isspace(*p))
			p++;
		while (isspace(*dp++ = *p++) == FALSE)
			;
		*--dp = '\0';
		terml = strlen(term);

		/* look for "> pathname" in command line */
		p = cmd;
		dp = path;
		while (*p != '>') {
			if (*p++ == '\0') {
				fprintf(stderr, ERRTO);
				exit(-1);
			}
		}
		while (isspace(*++p))
			;
		while (isspace(*dp++ = *p++) == FALSE)
			;
		*--dp = '\0';
#if MSDOS
		/* check for illegal DOS filenames */
		if (strncmp(path, "aux", 3) == SAME) {
			printf("filename conflicts with AUX:, ", path);
			sprintf(path, "xxx%d", filen++);
			printf("renamed %s\n", path);
		}
#endif
		while ((out = fopen(path, "wb")) == NULL) {
			printf("can't open %s, ", path);
			sprintf(path, "xxx%d", filen++);
			printf("renamed %s\n", path);
		}

		/* extract file contents according to sed/cat format */
		while (fgets(line, LINESIZE, in)) {
			if (sed) {
				if (strncmp(line, script, skip) == SAME)
					fputs(line + skip, out);
				else if (strncmp(line, term, terml) == SAME) {
					fclose(out);
					break;
				}
				else {
					fprintf(stderr, ELINE);
					break;
				}
			}
			else {
				if (strncmp(line, term, terml) == SAME) {
					fclose(out);
					break;
				}
				fputs(line, out);
			}
		}
	}
	fclose(in);
}
