/* ld.c    -  rudimentary loader for Small-C run-time
		support and compiler library
 */

/* 26-Jan-1989  v0.70 - A.J.Travis
 * 22-Jun-1991  v0.71 - J.G.Harston
 *   uses '/' instead of '_' in filenames
 *   main() ends with exit(0);
 */

#include <stdio.h>

#ifdef MSDOS
#define __BBC__ 0
#else
#define __BBC__ 1
#endif

#define SAME 0				/* strcmp() equal */
#define UNDEFINED -1			/* file status */
#define W_CAT 1			/* write catalogue info */
#define RW 3				/* read/write attributes */

char buf[BUFSIZ];			/* i/o buffer */
char *ofile;				/* output filename */
unsigned laddr;			/* load address */
unsigned eaddr;			/* execution address */
int in, out;				/* file descriptors */
int count;				/* number of bytes read/written */

int main(argc, argv)
int argc;
char *argv[];
{
	in = UNDEFINED;
	out = UNDEFINED;
	if (argc < 2)
		usage();
	ofile = "a/out";		/* default output file */
	laddr = 0x1902;			/* default load address */
	eaddr = 0x1902;			/* default exec address */
	while ((*argv[1] & 0xFF) == '-') {		/* BUG in compiler */
		if (strcmp(argv[1], "-o") == SAME) {
			if (argc < 3)
				usage();
			else {
				ofile = argv[2];
				argc = argc - 2;
				argv = argv + 2;
			}
		}
		else if (strcmp(argv[1], "-R") == SAME) {
			laddr = 0x8000;
			eaddr = 0x8000;
			--argc;
			argv++;
		}
		else
			usage();
	}
	if (argc < 2)
		usage();
	if ((out = creat(ofile, RW)) == -1) {
		fprintf(stderr, "ld: can't create %s\n", ofile);
		fatal(-1);
	}
	while (--argc) {
		if ((in = open(argv[1], 0)) == -1) {
			fprintf(stderr, "ld: can't open %s\n", argv[1]);
			fatal(-1);
		}
		while ((count = read(in, buf, BUFSIZ)) > 0)
			write(out, buf, count);
		close(in);
		argv++;
	}
	close(out);
#ifdef __BBC__
	attributes(ofile, laddr, eaddr);
#endif
	exit(0);
}

/*
 * report correct usage and exit
 */
usage()
{
	fprintf(stderr, "usage: ld [-o outfile] [-R] file1 ... [filen]\n");
	fatal(-1);
}

/*
 * close files and exit
 */
fatal(stat)
int stat;				/* exit status */
{
	if (in > 0)
		close(in);
	if (out > 0)
		close(out);
	exit(stat);
}

#ifdef __BBC__
/*
 * set load and execution attributes on object file
 */
attributes(file, load, exec)
char *file;				/* filename to set attributes on */
unsigned load;				/* load address */
unsigned exec;				/* execution address */
{
	int *fcb[9];			/* file control block */

	fcb[0] = 0;
	fcb[1] = load;
	fcb[2] = 0;
	fcb[3] = exec;
	fcb[4] = 0;
	fcb[5] = 0;
	fcb[6] = 0;
	fcb[7] = RW;
	fcb[8] = 0;
	return(osfile(file, fcb, W_CAT));
}
#endif
